"""
Interface for versioned services

Services that implement this interface can be managed by the VersionsService
"""
from abc import ABC
from typing import Optional


class IVersionedService(ABC):
    # The SemVer version of the service
    version: str

    # Whether the service supports the given protocol version
    def supports_protocol(self, protocol_version: float) -> bool:
        return self.version.startswith(str(protocol_version))


"""
Service for managing versions of services

The VersionsService class is a utility class for services which implement different versions of the protocol, and different versions of their
own service. For example, say you have a v1service, v11service, v2service and v21service with versions 1.0, 1.1, 2.0 and 2.1 respectively.
The VersionsService class can be used to manage these services:

from openframe_criteria_set_protocol.services import VersionsService


versions_service = new VersionsService([v1_service, v11_service, v2_service, v21_service]);
versions_service.get_latest_version(); // v21_service
versions_service.get_service_versions(1); // [v1_service, v11_service]
versions_service.get_service_versions(2); // [v2_service, v21_service]
versions_service.get('1.1'); // v11service
"""
class VersionsService:
    def __init__(self, versions: list[IVersionedService]):
        """
        :param versions: A list of services implementing IVersionedService
        """
        self.versions = versions

    def get_all(self) -> list[IVersionedService]:
        """
        Get all services
        """
        return self.versions

    def get_service_versions(self, protocol_version: float) -> list[IVersionedService]:
        """
        Retrieve the services which are compatible with the requested protocol version
        """
        return [service for service in self.versions if service.supports_protocol(protocol_version)]

    def get(self, version: str) -> Optional[IVersionedService]:
        """
        Retrieve a service by version
        """
        for service in self.versions:
            if service.version == version:
                return service

        return None

    def get_latest_version(self) -> IVersionedService:
        """
        Retrieve the latest service
        """
        return self.versions[-1]



import re
import unittest
from openframe_criteria_set_protocol.v1 import schemas as v1_schemas


class TestSchemas(unittest.TestCase):
    def test_criteria_set_id_schema(self):
        regex = re.compile(v1_schemas.criteria_set_id)

        self.assertFalse(regex.match('a b c'))
        self.assertFalse(regex.match('ææø'))
        self.assertFalse(regex.match('$a$b$c'))

        self.assertTrue(regex.match('a-b-c'))
        self.assertTrue(regex.match('a-b-c-1'))
        self.assertTrue(regex.match('a-b-c-1.2'))

    def test_version_schema(self):
        regex = re.compile(v1_schemas.version)

        self.assertFalse(regex.match('1'))
        self.assertFalse(regex.match('2.2'))
        self.assertFalse(regex.match('3.3.a'))

        self.assertTrue(regex.match('1.0.0'))



import datetime
import typing
from abc import ABC
from asyncio import StreamReader
from dataclasses import dataclass, field
from typing import Optional


@dataclass
class RgbColor:
    red: int
    green: int
    blue: int


Color = typing.Union[RgbColor, str]


@dataclass
class QualityStyle:
    primaryColor: Color
    secondaryColor: Color


@dataclass
class Quality:
    code: str
    style: Optional[QualityStyle] = None


@dataclass
class Metadata:
    id: str
    version: str
    date: datetime.datetime
    name: str
    description: str
    documentation: str
    qualities: Optional[list[Quality]] = None


@dataclass
class DocumentationItem(ABC):
    type: str
    label: str
    text: str
    url: Optional[str] = None


@dataclass
class PdfDocumentationItem(DocumentationItem):
    type: str = field(init=False, default='pdf')


@dataclass
class InlineDocumentationItem(DocumentationItem):
    type: str = field(init=False, default='text')


@dataclass
class LinkDocumentationItem(DocumentationItem):
    type: str = field(init=False, default='link')


TaskItemScalarValue = typing.Union[str, float, bool, None]
TaskItemValue = typing.Union[TaskItemScalarValue, list[TaskItemScalarValue]]

DefinitionType = typing.Literal['select-single', 'select-multiple', 'number', 'boolean']


@dataclass
class TaskItemValueMap(dict[str, TaskItemValue]):
    pass


@dataclass
class PointOption:
    label: str
    value: TaskItemScalarValue
    id: Optional[str] = None
    annotation: Optional[str] = None


@dataclass
class BaseTaskItemDefinition(ABC):
    type: DefinitionType


@dataclass
class SelectSingleType(BaseTaskItemDefinition):
    type: DefinitionType = field(init=False, default='select-single')
    options: list[PointOption]


@dataclass
class SelectMultipleType(BaseTaskItemDefinition):
    type: DefinitionType = field(init=False, default='select-multiple')
    options: list[PointOption]


@dataclass
class NumberType(BaseTaskItemDefinition):
    type: DefinitionType = field(init=False, default='number')
    minimum: Optional[float] = None
    maximum: Optional[float] = None
    step: Optional[float] = None


@dataclass
class BooleanType(BaseTaskItemDefinition):
    type: DefinitionType = field(init=False, default='boolean')


TaskItemDefinition = typing.Union[SelectSingleType, SelectMultipleType, NumberType, BooleanType]
CriteriaTreeElementType = typing.Literal['criterion', 'task-group', 'task', 'task-item']


@dataclass
class TaskItem:
    type: CriteriaTreeElementType = field(init=False, default='task-item')
    id: str
    definition: TaskItemDefinition
    label: Optional[str] = None
    tags: Optional[list] = None
    documentation: Optional[list[DocumentationItem]] = None
    description: Optional[str] = None
    annotation: Optional[str] = None
    providedData: Optional[dict[str, TaskItemValue]] = None
    data: Optional[dict[str, any]] = None


@dataclass
class Task:
    type: CriteriaTreeElementType = field(init=False, default='task')
    id: str
    title: str
    label: Optional[str] = None
    tags: Optional[list] = None
    documentation: Optional[list[DocumentationItem]] = None
    description: Optional[str] = None
    items: list[TaskItem] = field(default_factory=list)
    data: Optional[dict[str, any]] = None


@dataclass
class TaskGroup:
    type: CriteriaTreeElementType = field(init=False, default='task-group')
    id: str
    title: str
    label: Optional[str] = None
    tags: Optional[list] = None
    documentation: Optional[list[DocumentationItem]] = None
    items: list[Task] = field(default_factory=list)
    data: Optional[dict[str, any]] = None


@dataclass
class Criterion:
    type: CriteriaTreeElementType = field(init=False, default='criterion')
    id: str
    title: str
    quality: str
    label: Optional[str] = None
    tags: Optional[list] = None
    documentation: Optional[list[DocumentationItem]] = None
    items: list[TaskGroup] = field(default_factory=list)
    data: Optional[dict[str, any]] = None


@dataclass
class CriteriaTree:
    criteria: list[Criterion]
    result: Optional[dict[str, any]] = None


CriteriaTreeElement = typing.Union[Criterion, TaskGroup, Task, TaskItem]


@dataclass
class MetadataResponse:
    protocol: int = field(init=False, default=1)
    metadata: Metadata
    parameters: dict[str, typing.Any]


@dataclass
class CriteriaTreeResponse(CriteriaTree):
    pass


@dataclass
class StreamMatrixResponse:
    filename: str
    content_type: str
    stream: StreamReader



# Interface for services which implement the v1 protocol
from abc import ABC, abstractmethod
from typing import Optional, Any
from typing import TypeVar
from .types import CriteriaTree, Metadata, TaskItemValueMap, StreamMatrixResponse

ParametersType = TypeVar("ParametersType")


class IProtocolV1Service(ABC):

    # Specific version of this service, SemVer-formatted
    version: str

    # Get the criteria tree for the given criteria set ID and combination of parameters
    @abstractmethod
    def get_criteria_tree(
            self,
            criteria_set_id: str,
            parameters: ParametersType,
            values: Optional[TaskItemValueMap] = None,
            locale: Optional[str] = None
    ) -> CriteriaTree:
        pass

    # Validate the given parameters for the given criteria set ID
    @abstractmethod
    def validate_parameters(self, criteria_set_id: str, parameters: ParametersType):
        pass

    # Stream the matrix for the given criteria set ID, parameter combination and value map
    @abstractmethod
    def stream_matrix(
            self,
            criteria_set_id: str,
            parameters: ParametersType,
            values: Optional[TaskItemValueMap] = None,
            locale: Optional[str] = None
    ) -> StreamMatrixResponse:
        pass

    # Get the metadata of all criteria sets supported by this service
    @abstractmethod
    def get_criteria_set_metadata_list(self) -> list[Metadata]:
        pass

    # Get the JSON schema for the given criteria set ID
    @abstractmethod
    def get_parameters_json_schema(self, criteria_set_id: str) -> dict[str, Any]:
        pass
